from pydantic_settings import SettingsConfigDict

from rooms_shared_services.src.settings.dynamodb.base import BaseDynamodbSettings


class Settings(BaseDynamodbSettings):
    model_config = SettingsConfigDict(env_prefix="telegram_callbacks_")



import json
import logging
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Any
from uuid import UUID

from boto3.dynamodb.types import TypeDeserializer, TypeSerializer

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

type_serializer = TypeSerializer()
type_deserializer = TypeDeserializer()


def dump_value_dict(data_dict: dict, dynamodb_format: bool = False, json_format: bool = False) -> str | dict:
    """Convert a data dict to a dynamodb supported dict format or a json dump of this dict.

    Args:
        data_dict (dict): _description_
        dynamodb_format (bool): _description_. Defaults to False.
        json_format (bool): _description_. Defaults to False.

    Returns:
        str | dict: _description_
    """
    data_dict = {
        item_key: formatted_value
        for item_key, item_value in data_dict.items()
        if (formatted_value := convert_to_supported_format(item_value)) is not None
    }
    if dynamodb_format:
        data_dict = {item_key: serialize_value(item_value) for item_key, item_value in data_dict.items()}
    if json_format:
        return convert_to_json(data_dict)
    return data_dict


def load_value_json(value_json: str) -> dict:
    """Convert a json-serialized dict to a normal dict.

    Args:
        value_json (str): _description_

    Returns:
        dict: _description_
    """
    value_dict = json.loads(value_json)
    return parse_serialized_dict(value_dict=value_dict)


def parse_serialized_dict(value_dict: dict) -> dict:
    """Convert a dynamodb item value dict with dynamodb datatype symbols to a regular dict without datatype symbols.

    Args:
        value_dict (dict): _description_

    Returns:
        dict: _description_
    """
    value_dict = {item_key: type_deserializer.deserialize(item_value) for item_key, item_value in value_dict.items()}
    return {item_key: convert_from_supported_formats(item_value) for item_key, item_value in value_dict.items()}


def convert_from_supported_formats(item_value: Any) -> Any:
    """Convert a value in supported Dynamodb format to a more regular datatype format.

    Args:
        item_value (Any): _description_

    Returns:
        Any: _description_
    """
    match item_value:
        case Decimal():
            float_value = float(item_value)
            int_value = int(item_value)
            if float_value == int_value:
                converted_value = int_value
            else:
                converted_value = float_value
        case _:
            converted_value = item_value
    return converted_value


def convert_to_json(data_dict: dict) -> str:
    """Convert a data dict to a json str.

    Args:
        data_dict (dict): _description_

    Returns:
        str: _description_
    """
    return json.dumps(data_dict, indent=4)


def serialize_value(item_value: Any) -> dict:
    """Convert an item value to a dict with dynamodb supported datatype symbol as a key.

    Args:
        item_value (Any): _description_

    Returns:
        dict: _description_
    """
    return type_serializer.serialize(item_value)


def convert_to_supported_format(item_value: Any) -> str | list | Decimal | set | dict:
    """Convert an arbitrary value to a supported by DynamoDB format.

    Args:
        item_value (Any): _description_

    Returns:
        Any: _description_
    """
    match item_value:
        case Enum():
            formatted_value = item_value.value
        case UUID():
            formatted_value = str(item_value)
        case float():
            formatted_value = Decimal(str(item_value))
        case datetime():
            formatted_value = item_value.isoformat()
        case list():
            formatted_value = [convert_to_supported_format(item_value_elem) for item_value_elem in item_value]
        case set():
            formatted_value = {convert_to_supported_format(item_value_elem) for item_value_elem in item_value}
        case dict():
            formatted_value = {
                item_value_key: convert_to_supported_format(item_value_value)
                for item_value_key, item_value_value in item_value.items()
            }
        case _:
            formatted_value = item_value
    return formatted_value



from telebot import TeleBot
from telebot.types import InlineKeyboardMarkup
from telebot.util import quick_markup


def send_telegram_bot_message(
    bot: TeleBot,
    chat_id: int,
    text: str,
    reply_markup: InlineKeyboardMarkup | None = None,
) -> dict:
    """Send a message to a specified telegram bot chat.

    Args:
        bot (TeleBot): _description_
        chat_id (int): _description_
        text (str): _description_
        reply_markup (InlineKeyboardMarkup | None, optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """
    return bot.send_message(
        chat_id=chat_id,
        text=text,
        reply_markup=reply_markup,
    )


def get_inline_keyboard_markup(
    keyboard_labels: list[str],
    callback_values: list[str],
    row_width: int = 1,
) -> InlineKeyboardMarkup:
    """Get a markup for a telegram message inline keyboard.

    Args:
        keyboard_labels (list[str]): _description_
        callback_values (list[str]): _description_
        row_width (int): _description_. Defaults to 1.

    Returns:
        InlineKeyboardMarkup: _description_
    """
    data_options = [{"callback_data": callback_value} for callback_value in callback_values]
    markup_values = dict(zip(keyboard_labels, data_options))
    return quick_markup(values=markup_values, row_width=row_width)



from pydantic import BaseModel


class AbstractLambdaProcessorResponse(BaseModel):
    success: bool
    res: dict


class AbstractLambdaHandlerCallbackResponse(BaseModel):
    success: bool
    res: dict



import logging
from abc import ABC, abstractmethod

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class AbstractLambdaDataProcessor(ABC):
    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    @abstractmethod
    def run(self, *args, **kwargs):
        ...



import json
from decimal import Decimal


class RawDynamodbEncoder(json.JSONEncoder):
    def default(self, encode_obj):
        match encode_obj:
            case Decimal():
                return str(encode_obj)
        return super().default(encode_obj)



from abc import ABC, abstractmethod

from aws_lambda_powertools.utilities.typing import LambdaContext
from pydantic import BaseModel

from rooms_shared_services.src.lambda_handlers.api_gateway.abstract import AbstractLambdaDataProcessor
from rooms_shared_services.src.models.lmbda.http import HttpRequestContextData


class AbstractLambdaEventHandler(ABC):
    def __call__(self, event: dict, context: LambdaContext):
        event_model = self.parse_event(event)
        context_model = self.parse_context(context)
        try:
            run_result = self.run(event_model, context_model)
        except ValueError:
            return {
                "statusCode": 400,
            }
        return {"statusCode": 200, "headers": {"Content-Type": "*/*"}, "body": {"result": run_result}}

    @abstractmethod
    def parse_event(self, event: dict) -> BaseModel:
        ...

    def parse_context(self, context: LambdaContext) -> HttpRequestContextData:
        return HttpRequestContextData.model_validate(context.__dict__)

    @abstractmethod
    def provide_data_processor(self) -> AbstractLambdaDataProcessor:
        ...

    @abstractmethod
    def run(self, event_data, context_data) -> dict:
        ...



import logging
from typing import Literal, TypeVar

from pydantic import BaseModel

from rooms_shared_services.src.models.lmbda.abstract import AbstractLambdaEventModel
from rooms_shared_services.src.storage.models import BaseDynamodbModel

logger = logging.getLogger()
logger.setLevel(logging.INFO)


ImageVariant = Literal["new", "old"]
StreamViewType = Literal["KEYS_ONLY", "NEW_IMAGE", "OLD_IMAGE", "NEW_AND_OLD_IMAGES"]
DynamodbStreamModelType = TypeVar("DynamodbStreamModelType", bound=BaseDynamodbModel)


class DynamodbStreamRecord(BaseModel):
    keys: dict
    new_image: dict | None = None
    old_image: dict | None = None
    stream_view_type: str

    @classmethod
    def from_lambda_event(cls, record: dict):
        record_params = {
            "keys": record["Keys"],
            "new_image": record["NewImage"] if "NewImage" in record else None,
            "old_image": record["OldImage"] if "OldImage" in record else None,
            "stream_view_type": record["StreamViewType"],
        }
        return cls.model_validate(record_params)


class DynamodbStreamContextData(BaseModel):
    function_name: str
    function_version: str
    invoked_function_arn: str
    memory_limit_in_mb: str
    aws_request_id: str
    log_group_name: str
    log_stream_name: str
    identity: str | None = None
    client_context: str | None = None

    @classmethod
    def from_context_object(cls, context_obj):
        return cls.model_validate(context_obj.__dict__)


class DynamodbData(BaseModel):
    Keys: dict
    NewImage: dict | None
    OldImage: dict | None
    SequenceNumber: int
    SizeBytes: int
    StreamViewType: str

    def from_stream_image(
        self,
        image_variant: ImageVariant,
        model_type: type[DynamodbStreamModelType],
    ) -> DynamodbStreamModelType | None:
        match image_variant:
            case "new":
                item_image = self.NewImage
            case "old":
                item_image = self.OldImage
            case _:
                raise ValueError("Invalid image variant")
        if item_image is None:
            return None
        item_data = {**self.Keys, **item_image}
        return model_type.validate_dynamodb_item(data_dict=item_data, from_raw=True)


class DynamodbRecord(BaseModel):
    eventID: int
    eventName: str
    eventVersion: str
    eventSource: str  # ex: "aws:dynamodb"
    awsRegion: str  # "us-east-1"
    dynamodb: DynamodbData
    eventSourceARN: str  # ex: "stream-ARN"


class DynamodbStreamEventModel(AbstractLambdaEventModel):
    Records: list[DynamodbRecord]

    @classmethod
    def from_lambda_event(cls, event: dict):
        return cls.model_validate(event)



import os

import boto3
import yaml

from rooms_shared_services.src.settings.product_categories import Settings as ProductCategorySettings

script_dir = os.path.dirname(__file__)

CATEGORIES_PATH = "{}/categories.yml".format(script_dir)


class CategoryProvider(object):
    def __init__(self):
        """Set attributes."""
        self._settings = ProductCategorySettings()
        self.cat_list = []
        self.validated_cat_list = []
        self.cat_names = []
        self.not_allowed = [" "]
        self.root = ["categories"]
        self.s3_client = boto3.client("s3", region_name=self._settings.region_name)

    def __call__(self):
        self.provide_category_list()
        self.validate_cat_pairs()
        return self.validated_cat_list

    def load_source(self):
        s3_obj = self.s3_client.get_object(
            Bucket=self._settings.source_data_bucket,
            Key=self._settings.source_data_object,
        )
        return yaml.load(stream=s3_obj["Body"], Loader=yaml.Loader)  # noqa: S506

    def get_product_categories(self):
        with open(CATEGORIES_PATH, "r") as stream:
            return yaml.load(stream=stream, Loader=yaml.Loader)  # noqa: S506

    def provide_category_list(self):
        raw_cats = self.load_source()
        self.collect_cat_pairs(raw_cats)

    def collect_cat_pairs(self, raw_cats: dict):
        for parent, cat in raw_cats.items():
            match cat:
                case str():
                    self.cat_list.append((cat, parent))
                case dict():
                    for cat_key, _ in cat.items():
                        self.cat_list.append((cat_key, parent))  # noqa: WPS220
                    self.collect_cat_pairs(raw_cats=cat)
                case list():
                    for cat_name in cat:
                        self.cat_list.append((cat_name, parent))  # noqa: WPS220

    def validate_root(self, cat_elem: str):
        if cat_elem in self.root:
            return None
        return cat_elem

    def validate_cat_pairs(self):  # noqa: WPS231
        for cat_item in self.cat_list:
            for name in cat_item:
                for snippet in self.not_allowed:
                    if snippet in name:
                        raise ValueError("Not allowed snippet {} in {}".format(snippet, name))  # noqa: WPS220
            cat_child, cat_parent = cat_item
            if cat_child == "other":
                cat_child = "other_{}".format(cat_parent)
            if cat_child in self.cat_names:
                raise ValueError("Duplicated category name: {}".format(cat_child))
            self.cat_names.append(cat_child)
            new_cat_item = [self.validate_root(cat_elem) for cat_elem in (cat_child, cat_parent)]
            if new_cat_item != [None, None]:
                self.validated_cat_list.append((new_cat_item[0], new_cat_item[1]))



from abc import abstractmethod

from pydantic import BaseModel

from rooms_shared_services.src.lambda_handlers.abstract.handlers import AbstractNestedLambdaEventHandler
from rooms_shared_services.src.models.lmbda.abstract import AbstractLambdaEventModel
from rooms_shared_services.src.models.lmbda.dynamdb_stream import DynamodbStreamEventModel


class AbstractDynamodbStreamLambdaEventHandler(AbstractNestedLambdaEventHandler):
    """Handle Dynamodb Stream events."""

    @abstractmethod
    def provide_lambda_event_model(self, event: dict) -> AbstractLambdaEventModel:
        """Validate dynamodb stream lambda event model.

        Args:
            event (dict): _description_

        Returns:
            AbstractLambdaEventModel: _description_
        """
        ...

    def run_processor(self, data_model: BaseModel, response: dict) -> dict | None:
        return self.event_data_processor(data_model, response)


class BaseDynamodbStreamLambdaEventHandler(AbstractNestedLambdaEventHandler):
    def provide_data_model_array(self, event_model: AbstractLambdaEventModel):
        if not isinstance(event_model, DynamodbStreamEventModel):
            raise ValueError("Invalid event model type.")
        return DynamodbStreamEventModel.root



from abc import ABC, abstractmethod
from typing import TypeVar

from aws_lambda_powertools.utilities.typing import LambdaContext
from pydantic import BaseModel

from rooms_shared_services.src.lambda_handlers.abstract.callbacks import AbstractLambdaHandlerCallback
from rooms_shared_services.src.lambda_handlers.abstract.processors import AbstractLambdaDataProcessor
from rooms_shared_services.src.models.lmbda.abstract import AbstractLambdaEventModel
from rooms_shared_services.src.models.lmbda.responses import (
    AbstractLambdaHandlerCallbackResponse,
    AbstractLambdaProcessorResponse,
)

LambdaProcessorResponseType = TypeVar("LambdaProcessorResponseType", bound=AbstractLambdaProcessorResponse)
LambdaHandlerCallbackResponseType = TypeVar(
    "LambdaHandlerCallbackResponseType",
    bound=AbstractLambdaHandlerCallbackResponse,
)


# One of the handlers, that must handle a received lambda event.
class AbstractNestedLambdaEventHandler(ABC):
    label: str
    event_data_processor: AbstractLambdaDataProcessor
    callback: AbstractLambdaHandlerCallback
    response: list[dict]
    lambda_context: LambdaContext

    def __init__(self, lambda_context: LambdaContext, *args, **kwargs):
        """Set attributes.

        Parameters:
            lambda_context (LambdaContext): _description_
            args: _description_
            kwargs: _description_
        """
        self.lambda_context = lambda_context
        self.event_data_processor = self.provide_data_processor(*args, **kwargs)
        self.callback = self.provide_callback(*args, **kwargs)
        self.response_array: list = []

    def __call__(self, event_model: AbstractLambdaEventModel):
        self.handle_event_model(event_model)
        return self.response

    def provide_processor_params(self, data_model: BaseModel) -> dict:
        """Convert event_model into processor input params.

        Args:
            data_model (BaseModel): _description_

        Returns:
            dict: _description_
        """
        return data_model.model_dump()

    def handle_data_model(self, data_model: BaseModel | None):
        if data_model is None:
            response = {"event_match": False}
        else:
            response = {}
            self.run_all(data_model, response)
        self.response_array.append(response)

    def handle_event_model(self, event_model: AbstractLambdaEventModel):
        data_model_array = self.provide_data_model_array(event_model)
        for data_model in data_model_array:
            self.handle_data_model(data_model=data_model)

    def run_processor(self, data_model: BaseModel, response: dict) -> dict | None:
        processor_input_params = self.provide_processor_params(data_model=data_model)
        try:
            return self.event_data_processor(response=response, **processor_input_params)
        except Exception as err:
            response.update({"processor": {"error": str(err)}})
            return None

    def run_callback(self, res: dict, response: dict) -> None:
        if self.callback is None:
            response.update({"variant": "callback", "provided": "false"})
        try:
            self.callback(res, response)
        except Exception as err:
            response.update({"variant": "callback", "error": str(err)})

    def run_all(self, data_model: BaseModel, response: dict) -> None:
        processor_result = self.run_processor(data_model=data_model, response=response)
        if "error" in response["processor"]:
            response.update({"variant": "callback", "result": "None"})
        else:
            if processor_result is None:
                raise ValueError("Invalid processor result.")
            self.run_callback(res=processor_result, response=response)

    @abstractmethod
    def provide_data_model_array(self, event_model: AbstractLambdaEventModel):
        ...

    @abstractmethod
    def provide_data_processor(self, *args, **kwargs) -> AbstractLambdaDataProcessor:
        ...

    @abstractmethod
    def provide_callback(self, *args, **kwargs) -> AbstractLambdaHandlerCallback:
        ...


# A combined handler, that must trigger everyone of collected individual handlers.
class AbstractCombinedLambdaEventHandler(ABC):
    handler_types: list[type[AbstractNestedLambdaEventHandler]]

    def __call__(self, event: dict, lambda_context: LambdaContext):
        handler_collection = self.collect_handlers(lambda_context=lambda_context)
        event_model: AbstractLambdaEventModel = self.provide_lambda_event_model(event)
        try:
            handler_results = [lambda_handler(event_model) for lambda_handler in handler_collection]
        except Exception as err:
            return {"error": str(err)}

        return self.provide_lambda_response(handler_results)

    @abstractmethod
    def provide_lambda_event_model(self, event: dict) -> AbstractLambdaEventModel:
        """Validate lambda event model.

        Args:
            event (dict): _description_

        Returns:
            AbstractLambdaEventModel: _description_
        """
        ...

    def collect_handlers(self, lambda_context: LambdaContext) -> list[AbstractNestedLambdaEventHandler]:
        return [handler_type(lambda_context) for handler_type in self.handler_types]

    def provide_lambda_response(self, result_array_list: list[dict]):
        return {
            "handlers": [
                {"name": lambda_handler.label, "result_array": result_array}
                for lambda_handler, result_array in zip(self.handler_types, result_array_list)
            ],
        }



import logging

from aws_lambda_powertools.utilities.typing import LambdaContext

from rooms_shared_services.src.lambda_handlers.abstract.handlers import AbstractLambdaEventHandler
from rooms_shared_services.src.models.lmbda.http import HttpRequestContextData, HttpRequestEventData

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class APIGatewayRequestHandler(AbstractLambdaEventHandler):
    """Handle API GATEWAY requests."""

    def parse_event(self, event: dict) -> HttpRequestEventData:
        return HttpRequestEventData.from_lambda_event(event=event)

    def parse_context(self, context: LambdaContext) -> HttpRequestContextData:
        return HttpRequestContextData.model_validate(context.__dict__)

    def run(self, event_data: HttpRequestEventData, context_data: HttpRequestContextData) -> dict:
        data_processor = self.provide_data_processor()
        return data_processor(event_data, context_data)
